#!/usr/bin/env python

'''
biomorph.py -x <experiment>
   runs a neural network experiment.
biomorph.py -w <world>
   creates an ALife world of agents instead of an experiment (see world.py).

This module implements experiments with a neural network.

A Network consists of two or more Layers of units.

An Experiment has a Network and one more "conditions", each with a different Pattern or
Sequence Generator.

A pattern is a list of numbers for the input layer, and for supervised learning, a list
of target numbers for each output layer.

When an Experiment is run, it records error statistics.

When an Experiment is created, it is added to the dict EXPERIMENTS.

Assuming you are running this in a version of Python greater than or equal to 2.5
and this file is executable, you can run the program with graphics by typing
./biomorph.py -x <exp>
where "exp" is the name of one of the Experiments.  If it doesn't match any name,
the first Experiment in EXPERIMENTS is loaded.
'''

from pattern import *
from world import *

try:
   import psyco
   psyco.full()
except:
   pass

# Dict of Experiments
EXPERIMENTS = {}

# The world
WORLD = None

class Experiment(object):
    '''A Network with a set of conditions, each with a separate pattern generator.'''

    def __init__(self, name, network, conditions = None):
        '''Initialize name, network, learning rate, conditions, winners, error; add to EXPERIMENTS.'''
        self.network = network
        self.errors = []
        self.trials = 0
        # Index of current pattern generator
        self.condition = 0
        # Only applies to competitive learning
        self.winners = set([])
        self.current_error = 0.0
        self.conditions = conditions
        self.pat_gen = self.conditions[0]
        EXPERIMENTS[name] = self
        
    def step(self, train = True, showError = True, show_act = False):
        '''Run the Experiment on one pattern, return the target patter, error, # winners.'''
        pat = self.pat_gen()
        error = self.network.step(pat, train, show_act)
        if train:
            self.errors.append(error[0])
            self.trials += 1
        if show_act:
            self.show_target(pat)
        if showError:
            print 'Pat error'.ljust(12),
            print '%.3f' % error[0]
        return pat, error[0], error[1]

    def run(self, n, train = True):
        '''Run the Experiment on n patterns, training it and incrementing trials if train is True.'''
        self.current_error = 0.0
        self.winners = set([])
        trial0 = self.trials
        for i in range(n):
            pat_err_win = self.step(train, showError = False)
            self.current_error += pat_err_win[1]
            if pat_err_win[2]:
                # Competitive learning
                self.winners = self.winners | pat_err_win[2]
        print self.trials, 'trials'
        print 'Run error:',
        print '%.3f' % (self.current_error / n)
        if len(self.winners) > 0:
            print '# winners:', len(self.winners)

    def test(self, n):
        '''Test the network on n patterns.'''
        self.current_error = 0.0
        self.winners = set([])
        for i in range(n):
            pat_err_win = self.step(False, showError = True, show_act = True)
            self.current_error += pat_err_win[1]
            if pat_err_win[2]:
                # Competitive learning
                self.winners = self.winners | pat_err_win[2]
        print 'Run error'.ljust(12),
        print '%.3f' % (self.current_error / n)
        if len(self.winners) > 0:
            print '# winners:', len(self.winners)

    def reinit(self):
        '''Start from 0 trials and reinitalize weights in network.'''
        self.network.reinit()
        self.trials = 0

    def next_condition(self):
        '''Change to the next condition (if there is more than one).'''
        if len(self.conditions) > 1:
            self.condition = (self.condition + 1) % len(self.conditions)
            self.pat_gen = self.conditions[self.condition]
            self.reinit()
            print 'Changing to experiment condition', self.condition

    def show(self):
        '''Print activations for output layer of network.'''
        self.network.show()

    def show_target(self, pattern):
        '''Print the target for the pattern.'''
        print '['.rjust(12),
        for v in pattern[1]:
            print '%.3f' % v,
        print ']'

    def show_weights(self):
        self.network.show_weights()

#################
### Some examples
#################

### OR, no hidden layer
##or_pa_exp = Experiment('or', Network('or_pa', layers = [Layer('in', 2),
##                                                        Layer('out', 1)]),
##                       conditions = [or_pg])
##
### XOR, hidden layer
##xor_pa_exp = Experiment('xor',
##                        Network('xor_pa',
##                                layers = [Layer('in', 2),
##                                          Layer('hid', 2, lr = .2, bipolar=False),
##                                          Layer('out', 1, lr = .2, bipolar=False)]),
##                        conditions = [xor_pg])
##

## Auto-association
aa_42_exp = Experiment('aa42',
                      Network('aa', layers = [Layer('in', 4),
                                              Layer('hid', 2, lr=.2, bipolar=False),
                                              Layer('out', 4, lr=.2, bipolar=False)]),
                     conditions = [aa_4_pg])

aa_41_exp = Experiment('aa41',
                      Network('aa', layers = [Layer('in', 4),
                                              Layer('hid', 1, lr=.005, bipolar=False),
                                              Layer('out', 4, lr=.005, bipolar=False)]),
                      conditions = [aa_4_pg])

aa_6_exp = Experiment('aa6',
                      Network('aa', layers = [Layer('in', 6),
                                              Layer('hid', 2, lr=.01, bipolar=False),
                                              Layer('out', 6, lr=.01, bipolar=False)]),
                      conditions = [aa_6_pg])

# Competitive learning, 4 input units, 2 categories (and category units)
cl_4_2_exp = Experiment('cl',
                        CompetNetwork('cl',
                                      [Layer('in', 4), CompetLayer('out')]),
                        conditions = [cl_4_2_pg])

# Kohonen net, random inputs in the unit square.
koh1_exp = Experiment('koh1',
                      CompetNetwork('koh1',
                                    [Layer('in', 2),
                                     KohLayer('out', lr = .2, dimensions = (10, 10),
                                              sigma = .8, decay = .999)]),
                      conditions = [unit_square_pg])

# Kohonen net, random inputs in the unit square, clustered around .5.
koh1g_exp = Experiment('koh1g',
                       CompetNetwork('koh1g',
                                     [Layer('in', 2),
                                      KohLayer('out', lr = .2, dimensions = (10, 10),
                                               sigma = .8, decay = .999)]),
                       conditions = [unit_square_gauss_pg])

if __name__ == "__main__":
    if len(sys.argv) > 2:
        # Specify an experiment or a world file
        if sys.argv[1] == '-x':
            # An experiment
            # Do graphics
            root = Tk()
            # Find an Experiment with the name passed as the first argument
            experiment = EXPERIMENTS.get(sys.argv[2], None)
            if not experiment:
                if len(EXPERIMENTS) > 0:
                    experiment = EXPERIMENTS.values()[0]
            if experiment:
                NNFrame(root, experiment).mainloop()
        elif sys.argv[1] == '-w':
            # A world file
            frame = WorldFrame(Tk(), worldfile=sys.argv[2])
            WORLD = frame.canvas
            frame.mainloop()
        else:
            print 'Option', sys.argv[1], "doesn't make any sense!"
    else:
        # Don't run an experiment; just run the basic world
        frame = WorldFrame(Tk())
        WORLD = frame.canvas
        frame.mainloop()
