from utils import *
import thing
import world

# Labels for the different things sensed
SENSATIONS = { 
              # We only want the odor of plants for this world
              "odor": ("sweet",),     #  "pungent", "sweet", "gamey", "fragrant"),
              "texture": ("prickly", "hard", "fuzzy", "soft", "slippery"),
              "sound": ("ouch_front", "ouch_right", "ouch_left",
                        "mmm_front", "mmm_right", "mmm_left"),
              "color": ("red", "green")}

# Parameters for the kinds of senses
SMELL_REINF = -1
SMELL_RADIUS = 450
MAX_SMELLABLE = 4

TOUCH_REINF = -1
TOUCH_RADIUS = 24
TOUCH_ANGLE = 360
MAX_TOUCHABLE = 3

FEEL_REINF = -1
FEEL_ANGLE = 120
# FEEL_RADIUS = 42
FEEL_RADIUS = 18
FEELERS = 3
# FEELERS = 1

HEARING_REINF = -1
HEARING_RADIUS = 160
MAX_HEARABLE = 4

SIGHT_REINF = -1
SIGHT_RADIUS = 400
MAX_VISIBLE = 2

### Math we need here (and maybe elsewhere)

def get_dist(thing1, thing2):
    '''Integer Euclidian distance between centers of two things.'''
    return get_point_dist(thing1.x, thing1.y, thing2.x, thing2.y)

def get_angle(thing1, thing2):
    '''Angle in degrees from center of thing1 to center of thing2.'''
    return get_point_angle(thing1.x, thing1.y, thing2.x, thing2.y)

def get_heading_angle(thing1, thing2):
    '''Angle in degrees between heading of thing1 and line connecting thing1 and thing2.'''
    return (thing1.heading - get_angle(thing1, thing2)) % 360

def angle2feature(angle, section = 120):
    '''Convert an angle (degrees) to a string representing relative position.'''
    if 0 < angle < section/2 or 360 - section/2 < angle < 359:
        return 'front'
    elif section/2 < angle < section + section/2:
        return 'right'
    elif 360 - section - section/2 < angle < 360 - section/2:
        return 'left'
    else:
        return '?'

class Sense:
    '''One of the possibly multiple sense inputs for a Critter.'''

    reinforcement = 0

    def __init__(self, critter, feature, radius, angle = 360, color = "blue", max_sensible = 1,
                 positions = 1):
        '''Assign basic attributes, including the number of "positions".  Create the graphic(s).'''
        self.critter = critter
        self.radius = radius
        self.angle = angle
        self.set_start()
        self.world = critter.world
        self.feature = feature
        self.sensations = SENSATIONS[feature]
#        self.n_features = len(self.sensations) * positions
        self.set_positions(positions)
        self.set_n_features()
        self.max_sensible = max_sensible
        self.color = color
        # Number of input units required by network
        self.make_graphic()

    def set_positions(self, positions):
        '''Overridden in Feel.'''
        self.positions = positions

    def set_n_features(self):
        self.n_features = len(self.sensations) * self.positions

    def describe(self):
        '''Print out basic stuff.'''
        print "rad: " + str(self.radius),

    def make_graphic(self):
        '''Create the graphic that appears when senses are being shown.  Overridden in Feel.'''
        if self.world.show_senses:
            outline = self.color
        else:
            outline = ""
        rect = self.get_rect()
        if self.angle < 360:
            self.id = self.world.create_arc(rect[0], rect[1], rect[2], rect[3],
                                             start = self.critter.heading - self.angle / 2,
                                             outline = outline,
                                             extent = self.angle,
                                             width = 1)
        else:
            self.id = self.world.create_oval(rect[0], rect[1], rect[2], rect[3],
                                              outline = outline,
                                              width = 1)
        self.world.addtag_withtag("sense", self.id)
        self.world.tag_lower(self.id, self.critter.id)

    def set_start(self):
        '''Set the start angle.'''
        self.start = self.critter.heading - self.angle / 2

    def toggle_show(self, show):
        '''Show or unshow the graphic.'''
        if show:
            self.world.itemconfigure(self.id, outline = self.color)
        else:
            self.world.itemconfigure(self.id, outline = "")

    def get(self):
        '''What is sensed on a particular time step.'''
        return []

    def to_vector(self, sensed = None):
        '''A vector of values (0<v<1) corresponding to what is sensed, passed to network.'''
        return [float(x) / (self.max_sensible * self.critter.max_strength) for x in self.get_ints(sensed)]

    def get_ints(self, sensed=None):
        '''A list of sums of strengths for each attribute.'''
        if not sensed:
            sensed = self.get()
        ints = []
        for f in self.sensations:
            total = 0
            for s in sensed:
                if s[0] == f:
                    total += s[1]
            ints.append(total)
        return ints

    def get_walls(self):
        '''A list of pairs for each wall that overlaps with the Sense bounding box.'''
        return [("hard", 2000) for w in self.world.walls if w.overlaps(self)]

    def get_things(self):
        '''A list of pairs (feature, strength) of things that are sensed.'''
        nearest = self.nearest_within()
        return [(t.get_sensation(self.feature), t.strength) for t in self.nearest_within()\
                if t is not self.critter and not t.hidden]

    def within_range(self, t):
        '''Is Thing t is within the Sense radius?'''
        return self.within_dist(t)

    def nearness(self, t):
        '''Relative nearness of Thing t as a ratio of maximum distance.'''
        return 1.0 - (get_dist(self.critter, t) / float(self.radius))

    def nearest_within(self):
        '''Things within radius distance of center of critter.'''
        within = True
        found = []
        start = None
        checked = []
        while within:
            next = self.world.find_closest(self.critter.x, self.critter.y, halo = self.radius, start = start)
            if not next or next in checked:
                within = False
            else:
                checked.append(next)
#                print 'Next', next
                thing = self.world.things.get(next[0], None)
                if thing:
#                    print 'Next', next, 'thing', thing
                    if self.within_dist(thing):
                        found.append(thing)
            if next:
                start = next[0]
#        print 'Found', found
        return found

    def within_dist(self, t):
        '''Is distance from critter to Thing t less than the Sense radius?

        Note that distance is to *center* of t, not its edges.
        '''
        return get_dist(self.critter, t) < self.radius

    def within_angle(self, t):
        '''Is the critter within the Sense's angle?'''
        a = get_angle(self.critter, t)
        return self.start < a < self.start + self.angle\
               or self.start < a + 360 < self.start + self.angle

    def angle_to_wall(self, wall):
        '''The angle between critter's heading and the wall?'''
        heading = self.critter.heading
        if wall.vertical:
            if heading < 90 or heading > 300:
                return (90 - heading) % 360
            else:
                return (270 - heading) % 360
        elif heading < 90 or heading > 3000:
            return (0 - heading) % 360
        else:
            return (180 - heading) % 360

    def get_rect(self):
        '''Coordinates of rectangle around the Sense region.'''
        coords = (self.critter.x, self.critter.y)
        return (coords[0] - self.radius, coords[1] - self.radius,
                coords[0] + self.radius, coords[1] + self.radius)

    def turn(self):
        '''Adjust the graphic when the critter turns.'''
        if self.angle < 360:
            self.set_start()
            self.world.itemconfigure(self.id, start = self.start)

    def move(self):
        '''Adjust the graphic when the critter moves.'''
        rect = self.get_rect();
        self.world.coords(self.id, rect[0], rect[1], rect[2], rect[3])

    def die(self):
        '''Get rid of the canvas object.'''
        self.world.delete(self.id)

class Smell(Sense):
    '''Sense things (but not walls) within a radius, no information about direction or distance.'''

    def __init__(self, critter):
        self.reinforcement = SMELL_REINF
        self.max_dist = float(world.WIDTH)
# math.sqrt(world.WIDTH * world.WIDTH + world.HEIGHT * world.HEIGHT)
        Sense.__init__(self, critter, "odor", SMELL_RADIUS, color = "#0000AA",
                       max_sensible = MAX_SMELLABLE)

##    def nearest_within(self):
##        '''Things within radius distance of center of critter.'''
##        within = True
##        found = []
##        start = None
##        checked = []
##        while within:
##            next = self.world.find_closest(self.critter.x, self.critter.y, halo = self.radius, start = start)
##            if not next or next in checked:
##                within = False
##            else:
##                checked.append(next)
###                print 'Next', next
##                thing = self.world.things.get(next[0], None)
##                if thing:
###                    print 'Next', next, 'thing', thing
##                    if self.within_dist(thing):
##                        found.append(thing)
##            if next:
##                start = next[0]
###        print 'Found', found
##        return found

    def get_things(self):
        '''Overrides method in Sense.'''
        nearest = self.nearest_within()
        return [(t.get_sensation(self.feature), t.strength, self.intensity((t.x, t.y)), (t.x, t.y)) \
                for t in self.nearest_within() if t is not self.critter and not t.hidden]

    def get(self):
        '''The Things that are smelled.'''
        s = self.get_things()
        if len(s) > 1:
            s = sorted(s, cmp = lambda x,y: cmp(y[1], x[1]))
        return s[:self.max_sensible]

    def intensity(self, position):
        '''Intensity of a smell at position from the position of the critter.'''
        return intensity(self.scale_position(position),
                         self.scale_position((self.critter.x, self.critter.y)),
                         minimum=.05)

    def scale_position(self, position):
        '''Scale position to a fraction of maximum distance.'''
        return (position[0] / self.max_dist, position[1] / self.max_dist)

    def get_ints(self):
        '''Intensities of the smells from the position of the critter.'''
        things = self.get()
        odors = []
        for odor in self.sensations:
            o = 0.0
            for thing in things:
                if odor == thing[0]:
                    o += thing[2]
            o = min([1.0, o])
            odors.append(o)
        return odors

    def to_vector(self, sensed = None):
        '''A vector of values (0<v<1) corresponding to what is sensed, passed to network.'''
        return self.get_ints()

class Touch(Sense):
    '''Sense things right around the critter, no explicit information about direction or distance.'''

    def __init__(self, critter, positions = 1):
        self.reinforcement = TOUCH_REINF
        Sense.__init__(self, critter, "texture", TOUCH_RADIUS, angle = TOUCH_ANGLE, color = "#AAAA00",
                       max_sensible = MAX_TOUCHABLE, positions = positions)

    def get(self):
        '''Overrides method in Sense.'''
        s = self.get_things() + self.get_walls()
        if len(s) > 1:
            s = sorted(s, cmp = lambda x,y: cmp(y[1], x[1]))
        return s[:self.max_sensible]

    def get_things(self):
        '''Overrides method in Sense.'''
        nearest = self.nearest_within()
        return [(t.get_sensation(self.feature), t.strength, get_heading_angle(self.critter, t)) \
                for t in self.nearest_within() if t is not self.critter and not t.hidden]

    def to_vector(self, sensed=None):
        '''A vector of values for what is sensed, passed to network; overrides method in Sense (binary).'''
        return [(1.0 if x else 0.0) for x in self.get_ints(sensed)]

    def within_range(self, t):
        '''Overrides method in Sense.  Is Thing t within both distance and angle?'''
        return self.within_dist(t) and self.within_angle(t)

    def describe(self):
        '''Print out useful information.'''
        print "rad: " + str(self.radius) + ", start: " + str(self.start) + ", ang: " + str(self.angle),

class Feel(Sense):
    '''Feel textures at the end of feelers (or walls that they pass through).'''

    def __init__(self, critter, positions=[(45, 35), (0, 18), (5, 42), (315, 35)]):
        self.reinforcement = FEEL_REINF
        Sense.__init__(self, critter, "texture", FEEL_RADIUS, angle = FEEL_ANGLE, color = "#AA00AA",
                       max_sensible = positions, positions = positions)

    def set_positions(self, positions):
        '''Set angles of feelers.  Overrides method in Sense.'''
        self.positions = positions
##        if positions > 1:
##            left_max = self.angle / 2
##            angle_gap = self.angle / (positions - 1)
##            self.feeler_angles = [(left_max - (angle_gap * i)) % 360 for i in range(positions)]
##        else:
##            self.feeler_angles = [0]

    def set_n_features(self):
        self.n_features = len(self.sensations) * len(self.positions)

    def make_graphic(self):
        '''Make graphics for feelers.  Overrides method in Sense.'''
        if self.world.show_senses:
            fill = self.color
        else:
            fill = ""
        self.ids = []
        for a, l in self.positions:   #self.feeler_angles:
            coords = self.feeler_coords(a, l)
            new_id = self.world.create_line(coords[0], coords[1], coords[2], coords[3],
                                            fill = fill)
            self.ids.append(new_id)
            self.world.addtag_withtag("sense", new_id)

    def feeler_coords(self, angle, length):
        '''Coordinates of feel with angle.'''
        end_x, end_y = get_endpoint(self.critter.x, self.critter.y,
                                    (self.critter.heading + angle) % 360, length)
        return self.critter.x, self.critter.y, end_x, end_y

    def toggle_show(self, show):
        '''Show or unshow graphics.  Overrides method in Sense.'''
        for i in self.ids:
            if show:
                self.world.itemconfigure(i, fill = self.color)
            else:
                self.world.itemconfigure(i, fill = "")

    def get_felt(self):
        '''List of lists of Things and Walls felt by the feelers.'''
        found = []
        wall_bounds = self.world.vert_real_wall_bounds + self.world.horiz_real_wall_bounds
        for feeler_id in self.ids:
            coords = self.world.coords(feeler_id)
            found1 = []
            start = None
            close_enough = True
            if self.wall_felt(feeler_id, wall_bounds):
                found1.append('hard')
            while close_enough:
                next = self.world.find_closest(coords[2], coords[3], halo = 1, start = start)
                thing = self.world.things.get(next[0], None)
                if thing:
                    found1.append(thing.get_sensation(self.feature))
                    start = next[0]
                else:
                    close_enough = False
            found.append(found1)
        return found

    def wall_felt(self, feeler_id, walls = []):
        '''Is a wall felt by this feeler?'''
        # borders of walls, each a list of 4 coordinates
        wall_bounds = walls if walls else self.world.vert_real_wall_bounds + self.world.horiz_real_wall_bounds
        coords = self.world.coords(feeler_id)
        # start and end coordinates of feeler
        start_x, start_y, end_x, end_y = coords[0], coords[1], coords[2], coords[3]
        for w in wall_bounds:
            # end of feeler between y bounds of wall?
            end_between_y = w[1] < end_y < w[3]
            # start of feeler between y bounds of wall?
            start_between_y = w[1] < start_y < w[3]
            # one end of feeler between top and bottom?
            if end_between_y or start_between_y:
                # start and end are on opposite sides of the wall (above and below)?
                if end_x <= w[2] < start_x or end_x >= w[0] > start_x:
                    # start and end of feeler are both between top and bottom?
                    if end_between_y and start_between_y:
                        return True
                    # end is in the wall?
                    elif end_between_y and w[0] <= end_x <= w[2]:
                        return True
                    else:
                        # crosses vertical wall near end?
                        width = w[2] - w[0]
                        if w[3] - w[1] > width:
                            # has to be a vertical wall
                            wall_mid = w[0] + width / 2
                            x_ratio = abs(end_x - wall_mid) / abs(end_x - start_x)
                            y_diff = x_ratio * abs(end_y - start_y)
                            y_intersect = end_y + (y_diff if end_y < start_y else -y_diff)
                            if w[1] < y_intersect < w[3]:
                                return True
            # end of feeler between x bounds of wall?
            end_between_x = w[0] < end_x < w[2]
            # start of feeler between x bounds of wall?
            start_between_x = w[0] < start_x < w[2]
            # one end of feeler between left and right?
            if end_between_x or start_between_x:
                # start and end are on opposite sides of the wall (above and below)?
                if end_y <= w[3] < start_y or end_y >= w[1] > start_y:
                    # start and end of feeler are both between left and right?
                    if end_between_x and start_between_x:
                        return True
                    # end is in the wall?
                    elif end_between_x and w[1] <= end_y <= w[3]:
                        return True
                    else:
                        # crosses horizontal wall near end?
                        height = w[3] - w[1]
                        if w[2] - w[0] > height:
                            # has to be a horizontal wall
                            wall_mid = w[1] + height / 2
                            y_ratio = abs(end_y - wall_mid) / abs(end_y - start_y)
                            x_diff = y_ratio * abs(end_x - start_x)
                            x_intersect = end_x + (x_diff if end_x < start_x else -x_diff)
                            if w[0] < x_intersect < w[2]:
                                return True
        return False

    def die(self):
        '''Get rid of the graphics.  Overrides method in Sense.'''
        for i in self.ids:
            self.world.delete(i)

    def turn(self):
        '''Adjust graphics when critter turns.  Overrides method in Sense.'''
        for i, a_l in zip(self.ids, self.positions):
            coords = self.feeler_coords(a_l[0], a_l[1])
            self.world.coords(i, coords[0], coords[1], coords[2], coords[3])

    def move(self):
        '''Adjust graphics when critter moves.   Overrides method in Sense.'''
        for i, a_l in zip(self.ids, self.positions):
            coords = self.feeler_coords(a_l[0], a_l[1])
            self.world.coords(i, coords[0], coords[1], coords[2], coords[3])

    def get(self):
        '''Overrides method in Sense.  List of lists of features.'''
        return self.get_felt()

    def get_ints(self, sensed = None):
        '''A list of sums of strengths for each attribute in each feeler.'''
        if not sensed:
            sensed = self.get()
        ints = []
        for feeler_sensed in sensed:
            ints1 = []
            for f in self.sensations:
                total = 0
                for s in feeler_sensed:
                    if s == f:
                        total += 1
                ints1.append(total)
            ints.append(ints1)
        return reduce_lists(ints)

    def to_vector(self, sensed = None):
        '''Vector of values for what is sensed, passed to network.  Overrides method in Sense (binary.'''
        return [(1.0 if x else 0.0) for x in self.get_ints(sensed)]

    def describe(self):
        '''Print a little information.'''
        print "# feelers: ", self.positions,

class Sight(Sense):
    '''Sense lights.  Provides information about intensity but not direction.'''

    colors = ('red', 'green')

    def __init__(self, critter):
        self.reinforcement = SIGHT_REINF
        self.max_dist = float(world.WIDTH)
# math.sqrt(world.WIDTH * world.WIDTH + world.HEIGHT * world.HEIGHT)
        Sense.__init__(self, critter, 'color', SIGHT_RADIUS, color = "#AA0000",
                       max_sensible = MAX_VISIBLE)

    def intensity(self, position, brightness):
        '''Intensity of a light at position from the position of the critter.'''
        return intensity(self.scale_position(position),
                         self.scale_position((self.critter.x, self.critter.y)),
                         minimum=.05,
                         magnitude=brightness)

    def scale_position(self, position):
        '''Scale position to a fraction of maximum distance.'''
        return (position[0] / self.max_dist, position[1] / self.max_dist)

    def get_ints(self):
        '''Intensities of the lights from the position of the critter.'''
        return [self.intensity((light.x, light.y), light.brightness) for light in self.world.lights]

    def to_vector(self, sensed = None):
        '''A vector of values (0<v<1) corresponding to what is sensed, passed to network.'''
        return self.get_ints()
#    [x / self.max_dist for x in self.get_ints()]

    def toggle_show(self, show):
        """Don't bother trying to show Sight."""
        pass

class Hearing(Sense):
    '''Sense sounds.  Provides information about direction.'''

    def __init__(self, critter):
        self.reinforcement = HEARING_REINF
        Sense.__init__(self, critter, "sound", HEARING_RADIUS, color = "#AA0000",
                       max_sensible = MAX_HEARABLE)

    def get_things(self):
        '''Creates features that combine the sound with the position relative to ME.'''
        things = [t for t in self.world.sounds\
                  if self.within_range(t) and not self.critter.within([t.x, t.y])]
        return [(t.sensations[self.feature] + '_' + angle2feature(get_heading_angle(self.critter, t)), 1000) \
                for t in things]

    def get(self):
        '''Features combining sounds with their directions.'''
        s = self.get_things()
#        print "Hearing gets: " + str(s)
        if len(s) > 1:
            random.shuffle(s)
        return s[:self.max_sensible]
